unit untHttp;

interface

uses
  Classes,
  IdThread, IdHttp, IdComponent,
  untStreamTool, SyncObjs;

type

  TAsyncHttp = class(TIdThread)
  private
    FURL           : string;
    FHttp          : TIdHttp;
    FWriteEvent    : TMemoryStreamEx;
    FBufferReader  : TStreamReader;
    FReadPosition  : Integer;
    FOnReceived    : TNotifyEvent;
    FReceivedLines : TStringList;
    FStartRange    : Integer;
    FLastModified  : string;
    FOnStatus      : TIdStatusEvent;
    FStatus        : TIdStatus;
    FStatusText    : String;
    FOnComplete    : TNotifyEvent;
    FContentLength : Integer;
    FUserAgent     : string;
    FProxyPort     : Integer;
    FProxyHost     : string;
    FInited        : Boolean;
    FAddHeaders    : string;
    FResponseCode  : Integer;
    FBuffer        : TMemoryStream;
    procedure HttpReceived(const Buff; Count : int64);
    procedure HttpStatus(ASender: TObject; const AStatus: TIdStatus; const AStatusText: string);
    procedure SetOnReceived(const Value: TNotifyEvent);
    procedure SetStartRange(const Value: Integer);
    procedure SetLastModified(const Value: string);
    procedure RaiseReceivedEvent;
    procedure RaiseStatusEvent;
    procedure SetOnStatus(const Value: TIdStatusEvent);
    function  GetReplayCode : Integer;
    procedure SetOnComplete(const Value: TNotifyEvent);
    procedure SetContentLength(const Value: Integer);
    procedure SetUserAgent(const Value: string);
    procedure SetProxyHost(const Value: string);
    procedure SetProxyPort(const Value: Integer);
  public
    property    OnReceived : TNotifyEvent read FOnReceived write SetOnReceived;
    property    OnStatus: TIdStatusEvent read FOnStatus write SetOnStatus;
    property    OnComplete : TNotifyEvent read FOnComplete write SetOnComplete;
    property    ReceivedLines : TStringList read FReceivedLines;
    property    StartRange : Integer read FStartRange write SetStartRange;
    property    LastModified : string read FLastModified write SetLastModified;
    property    UserAgent    : string read FUserAgent write SetUserAgent;
    property    ContentLength : Integer read FContentLength write SetContentLength;
    property    ResponseCode : Integer read GetReplayCode;
    property    ProxyHost : string read FProxyHost write SetProxyHost;
    property    ProxyPort : Integer read FProxyPort write SetProxyPort;
    constructor Create(ACreateSuspended: Boolean = True); override;
    destructor  Destroy; override;
    procedure   BeforeRun; override;
    procedure   Run; override;
    procedure   AfterRun; override;
    procedure   AddHeader(HeaderName, Value : string);
    procedure   Get(URL : string);
  end;

implementation

{ TAsyncHttp }

procedure TAsyncHttp.SetOnComplete(const Value: TNotifyEvent);
begin
  FOnComplete := Value;
end;

procedure TAsyncHttp.SetContentLength(const Value: Integer);
begin
  FContentLength := Value;
end;

procedure TAsyncHttp.AddHeader(HeaderName, Value: string);
begin
  FAddHeaders := FAddHeaders + HeaderName + ': ' + Value + #13#10;
end;

procedure TAsyncHttp.SetUserAgent(const Value: string);
begin
  FUserAgent := Value;
end;

procedure TAsyncHttp.SetProxyHost(const Value: string);
begin
  FProxyHost := Value;
end;

procedure TAsyncHttp.SetProxyPort(const Value: Integer);
begin
  FProxyPort := Value;
end;

procedure TAsyncHttp.SetOnReceived(const Value: TNotifyEvent);
begin
  FOnReceived := Value;
end;

constructor TAsyncHttp.Create(ACreateSuspended: Boolean = True);
begin
  inherited Create(ACreateSuspended);

end;

destructor TAsyncHttp.Destroy;
begin

  inherited;
end;

procedure TAsyncHttp.Get(URL: string);
begin

  FURL := URL;
  self.Start;

end;

procedure TAsyncHttp.BeforeRun;
begin
  inherited;

end;

procedure TAsyncHttp.Run;
var
  headers : TStringList;
  I       : Integer;
begin
  inherited;

  FReceivedLines := TStringList.Create;

  FWriteEvent := TMemoryStreamEx.Create;
  FWriteEvent.OnWrite := HttpReceived;
  FBuffer := TMemoryStream.Create;
  FBufferReader       := TStreamReader.Create(FBuffer);
  FReadPosition       := 0;

  {*debug Config.SetReadProxy(FHttp)
  FHttp.Request.UserAgent   := gUserAgent;
  FHttp.Request.RawHeaders.Add('X-2ch-UA: ' + APP_2chUA);
  }

  FHttp := TIdHttp.Create(nil);
  FHttp.Request.ContentRangeStart := FStartRange;
  FHttp.Request.Connection := 'close';
  FHttp.Request.UserAgent := FUserAgent;

  headers := TStringList.Create;
  headers.Text := FAddHeaders;
  for I := 0 to headers.Count - 1 do
    FHttp.Request.RawHeaders.Add(headers[I]);

  if FProxyHost <> '' then
  begin
    FHttp.ProxyParams.ProxyServer := FProxyHost;
    FHttp.ProxyParams.ProxyPort   := FProxyPort;
    FHttp.Request.Pragma := 'no-cache';
  end;

  FHttp.Get(FURL, FWriteEvent);
  FHttp.Disconnect;

  FResponseCode  := FHttp.ResponseCode;
  FLastModified  := FHttp.Response.RawHeaders.Values['Last-Modified'];
  FContentLength := FHttp.Response.ContentLength;

  if Assigned(FOnComplete) then FOnComplete(self);

  FBuffer.Free;
  FWriteEvent.Free;
  FBufferReader.Free;

  FReceivedLines.Free;
  FHttp.Free;

  FURL := '';
  FReadPosition := 0;
  FOnReceived   := nil;
  FStartRange   := 0;
  FLastModified := '';
  FOnStatus     := nil;
//  FStatus       := 0;
  FStatusText   := '';
  FOnComplete   := nil;
  FContentLength := 0;
  FUserAgent     :='';
  FProxyPort     := 0;
  FProxyHost     := '';
  FInited        := false;
  FAddHeaders    := '';
  //FReceivedLines.Clear;

  Stop;

end;

procedure TAsyncHttp.AfterRun;
begin
  inherited;

end;

procedure TAsyncHttp.HttpReceived(const Buff; Count: int64);
var
  line : string;
begin

  line := '';
  FBuffer.Seek(0, soFromEnd);
  FBuffer.Write(Buff, Count);
  FBuffer.Seek(FReadPosition, soFromBeginning	);

  // ͊̕Oɏo\
  while FBufferReader.ReadLine(line) do
  begin
    FReadPosition := FBuffer.Position;
    FReceivedLines.Add(line);
  end;

  Synchronize(RaiseReceivedEvent);

end;

procedure TAsyncHttp.SetStartRange(const Value: Integer);
begin
  FStartRange := Value;
end;

procedure TAsyncHttp.SetLastModified(const Value: string);
begin
  FLastModified := Value;
end;

procedure TAsyncHttp.RaiseReceivedEvent;
begin
  if Assigned(FOnReceived) then
    FOnReceived(self)
end;

procedure TAsyncHttp.SetOnStatus(const Value: TIdStatusEvent);
begin
  FOnStatus := Value;
end;

procedure TAsyncHttp.HttpStatus(ASender: TObject; const AStatus: TIdStatus;
  const AStatusText: string);
begin

  FStatus     := AStatus;
  FStatusText := AStatusText;

  Synchronize(RaiseStatusEvent);

end;

procedure TAsyncHttp.RaiseStatusEvent;
begin

  if Assigned(FOnStatus) then
    FOnStatus(self, FStatus, FStatusText)

end;

function TAsyncHttp.GetReplayCode: Integer;
begin

  result := FResponseCode;

end;

end.
